@@ -1,7 +1,7 @@
 /* Simple box operations */
 
 /* 
- * Copyright (C) 2005 Elijah Newren
+ * Copyright (C) 2005, 2006 Elijah Newren
  * [meta_rectangle_intersect() is copyright the GTK+ Team according to Havoc,
  * see gdkrectangle.c.  As far as Havoc knows, he probably wrote
  * meta_rectangle_equal(), and I'm guessing it's (C) Red Hat.  So...]
@@ -55,11 +55,12 @@ meta_rectangle_region_to_string (GList      *region,
    */
   char rect_string[RECT_LENGTH];
 
+  GList *tmp = region;
+  char *cur = output;
+
   if (region == NULL)
     snprintf (output, 10, "(EMPTY)");
 
-  char *cur = output;
-  GList *tmp = region;
   while (tmp)
     {
       MetaRectangle *rect = tmp->data;
@@ -107,11 +108,12 @@ meta_rectangle_edge_list_to_string (GList      *edge_list,
    */
   char rect_string[EDGE_LENGTH];
 
+  char *cur = output;
+  GList *tmp = edge_list;
+
   if (edge_list == NULL)
     snprintf (output, 10, "(EMPTY)");
 
-  char *cur = output;
-  GList *tmp = edge_list;
   while (tmp)
     {
       MetaEdge      *edge = tmp->data;
@@ -620,15 +622,38 @@ meta_rectangle_expand_region (GList     *region,
                               const int  top_expand,
                               const int  bottom_expand)
 {
-  /* Now it's time to do the directional expansion */
+  return meta_rectangle_expand_region_conditionally (region,
+                                                     left_expand,
+                                                     right_expand,
+                                                     top_expand,
+                                                     bottom_expand,
+                                                     0,
+                                                     0);
+}
+
+GList*
+meta_rectangle_expand_region_conditionally (GList     *region,
+                                            const int  left_expand,
+                                            const int  right_expand,
+                                            const int  top_expand,
+                                            const int  bottom_expand,
+                                            const int  min_x,
+                                            const int  min_y)
+{
   GList *tmp_list = region;
   while (tmp_list)
     {
       MetaRectangle *rect = (MetaRectangle*) tmp_list->data;
-      rect->x      -= left_expand;
-      rect->width  += (left_expand + right_expand);
-      rect->y      -= top_expand;
-      rect->height += (top_expand + bottom_expand);
+      if (rect->width >= min_x)
+        {
+          rect->x      -= left_expand;
+          rect->width  += (left_expand + right_expand);
+        }
+      if (rect->height >= min_y)
+        {
+          rect->y      -= top_expand;
+          rect->height += (top_expand + bottom_expand);
+        }
       tmp_list = tmp_list->next;
     }
 
@@ -1010,6 +1035,30 @@ meta_rectangle_find_linepoint_closest_to_point (double x1,
 /*                                                                         */
 /***************************************************************************/
 
+gboolean
+meta_rectangle_edge_aligns (const MetaRectangle *rect, const MetaEdge *edge)
+{
+  /* The reason for the usage of <= below instead of < is because we are
+   * interested in in-the-way-or-adject'ness.  So, a left (i.e. vertical
+   * edge) occupying y positions 0-9 (which has a y of 0 and a height of
+   * 10) and a rectangle with top at y=10 would be considered to "align" by
+   * this function.
+   */
+  switch (edge->side_type)
+    {
+    case META_DIRECTION_LEFT:
+    case META_DIRECTION_RIGHT:
+      return BOX_TOP (*rect)      <= BOX_BOTTOM (edge->rect) &&
+             BOX_TOP (edge->rect) <= BOX_BOTTOM (*rect);
+    case META_DIRECTION_TOP:
+    case META_DIRECTION_BOTTOM:
+      return BOX_LEFT (*rect)      <= BOX_RIGHT (edge->rect) &&
+             BOX_LEFT (edge->rect) <= BOX_RIGHT (*rect);
+    }
+
+  g_assert_not_reached ();
+}
+
 static GList*
 get_rect_minus_overlap (const GList   *rect_in_list, 
                         MetaRectangle *overlap)
@@ -1065,8 +1114,7 @@ replace_rect_with_list (GList *old_element,
   if (!new_list)
     {
       /* If there is no new list, just remove the old_element */
-      ret = old_element->next;
-      g_list_remove_link (old_element, old_element);
+      ret = g_list_remove_link (old_element, old_element);
     }
   else
     {
@@ -1173,6 +1221,49 @@ get_disjoint_strut_list_in_region (const GSList        *old_struts,
   return struts;
 }
 
+gint
+meta_rectangle_edge_cmp_ignore_type (gconstpointer a, gconstpointer b)
+{
+  const MetaEdge *a_edge_rect = (gconstpointer) a;
+  const MetaEdge *b_edge_rect = (gconstpointer) b;
+  int a_compare, b_compare;
+
+  /* Edges must be both vertical or both horizontal, or it doesn't make
+   * sense to compare them.
+   */
+  g_assert ((a_edge_rect->rect.width  == 0 && b_edge_rect->rect.width == 0) ||
+            (a_edge_rect->rect.height == 0 && b_edge_rect->rect.height == 0));
+
+  a_compare = b_compare = 0;  /* gcc-3.4.2 sucks at figuring initialized'ness */
+
+  if (a_edge_rect->side_type == META_DIRECTION_LEFT ||
+      a_edge_rect->side_type == META_DIRECTION_RIGHT)
+    {
+      a_compare = a_edge_rect->rect.x;
+      b_compare = b_edge_rect->rect.x;
+      if (a_compare == b_compare)
+        {
+          a_compare = a_edge_rect->rect.y;
+          b_compare = b_edge_rect->rect.y;
+        }
+    }
+  else if (a_edge_rect->side_type == META_DIRECTION_TOP ||
+           a_edge_rect->side_type == META_DIRECTION_BOTTOM)
+    {
+      a_compare = a_edge_rect->rect.y;
+      b_compare = b_edge_rect->rect.y;
+      if (a_compare == b_compare)
+        {
+          a_compare = a_edge_rect->rect.x;
+          b_compare = b_edge_rect->rect.x;
+        }
+    }
+  else
+    g_assert ("Some idiot wanted to sort sides of different types.\n");
+
+  return a_compare - b_compare; /* positive value denotes a > b ... */
+}
+
 /* To make things easily testable, provide a nice way of sorting edges */
 gint
 meta_rectangle_edge_cmp (gconstpointer a, gconstpointer b)
@@ -1186,32 +1277,7 @@ meta_rectangle_edge_cmp (gconstpointer a, gconstpointer b)
   b_compare = b_edge_rect->side_type;
 
   if (a_compare == b_compare)
-    {
-      if (a_edge_rect->side_type == META_DIRECTION_LEFT ||
-          a_edge_rect->side_type == META_DIRECTION_RIGHT)
-        {
-          a_compare = a_edge_rect->rect.x;
-          b_compare = b_edge_rect->rect.x;
-          if (a_compare == b_compare)
-            {
-              a_compare = a_edge_rect->rect.y;
-              b_compare = b_edge_rect->rect.y;
-            }
-        }
-      else if (a_edge_rect->side_type == META_DIRECTION_TOP ||
-               a_edge_rect->side_type == META_DIRECTION_BOTTOM)
-        {
-          a_compare = a_edge_rect->rect.y;
-          b_compare = b_edge_rect->rect.y;
-          if (a_compare == b_compare)
-            {
-              a_compare = a_edge_rect->rect.x;
-              b_compare = b_edge_rect->rect.x;
-            }
-        }
-      else
-        g_assert ("Some idiot wanted to sort sides of different types.\n");
-    }
+    return meta_rectangle_edge_cmp_ignore_type (a, b);
 
   return a_compare - b_compare; /* positive value denotes a > b ... */
 }
@@ -1453,11 +1519,12 @@ fix_up_edges (MetaRectangle *strut,        MetaEdge *edge,
           /* If this is the edge that overlaps, then we need to split it */
           if (edges_overlap (cur, &overlap))
             {
+              GList *delete_me = tmp;
+
               /* Split this edge into some new ones */
               *strut_edges = split_edge (*strut_edges, cur, &overlap);
 
               /* Delete the old one */
-              GList *delete_me = tmp;
               tmp = tmp->next;
               g_free (cur);
               *strut_edges = g_list_delete_link (*strut_edges, delete_me);
